package keskustelubotti;

import java.io.*;

/**
 * Itse keskustelubotti, kynnist oman threadin bottia varten ja pyrii nukkumissilmukassa,
 * josta se her jos on aika puhua tai ksitell uusi botille sanottu lause.
 *
 * Kytt lauseidenksittelyluokkia asioidensa hoitamiseen.
 *
 * @author Samuli Kaipiainen
 * @version 0.1
 */

public class Botti implements Runnable
{
	private Thread botti;
	private BottiUI emo;
	private Asetukset asetukset;
	private Sanavarasto sanavarasto;
	private Lauseparseri parseri;
	private Lausegeneroija generoija;
	private Lausejono jonoBotille;
	private Lausejono jonoBotilta;

	/**
	 * Luo uuden botin; luo botti-threadin joka siirt valmiita lauseita
	 * lausejonosta emoikkunan (tai muun BottiUI:n) sano-metodille
	 *
	 * @param emo BottiUI-kyttliittym, jonka sano- ja ajattele-metodeita botti kutsuu
	 * @param asetukset Botin (ja muiden) asetukset
	 * @param sanavarasto Botin (ja muiden) kyttm sanavarasto
	 */
	public Botti(BottiUI emo, Asetukset asetukset, Sanavarasto sanavarasto)
	{
		// luodaan apuoliot
		this.emo = emo;
		this.asetukset = asetukset;
		this.sanavarasto = sanavarasto;
		this.parseri = new Lauseparseri(sanavarasto, asetukset);
		this.generoija = new Lausegeneroija(sanavarasto, asetukset);
		this.jonoBotille = new Lausejono();
		this.jonoBotilta = new Lausejono();

		// murahdetaan tyhjlle sanavarastolle, ettei tule ikvi null pointer exceptioneja
		if (sanavarasto.onTyhj()) sanavarasto.lisSana("MUR");

		// luodaan bottithread
		botti = new Thread(this);
		botti.start();

//		emo.ajattele(asetukset.bottiNimi + " kynniss");
	}

	/**
	 * Kutsutaan kun luodaan uusi botti-thread, pyrii koko ajan synkronoidussa wait-silmukassa
	 * ja kutsuu tarvittaessa lauseidenksittelymetodeita
	 */
	public void run()
	{
		// seuraavaksi sanottavan lauseen ajastettu aika
		long nukkumisAika;

		// tervehdys
		sanoSatunnainenLause();

		// listn heti satunnainen lause joskus sanottavaksi; pidetn botilta tulevien
		// lauseiden jonossa koko ajan vhintn yksi lause (oma-aloitteinen)
		lisOmaLause();

		while (true) {
			// kauanko saadaan nukkua
			nukkumisAika = jonoBotilta.seuraavanAika() - System.currentTimeMillis();

			// nukutaan jos saadaan, muuten on aika sanoa seuraava lause
			if (nukkumisAika > 0) {
				try {
					synchronized (this) { wait(nukkumisAika); }
				} catch (InterruptedException e) {}
			} else {
				// sanotaan seuraava lause valmiiden jonosta
				sanoSeuraavaLause();

				// jos ei ole lis lauseita (vastauksia) jonossa, listn oma-aloitteinen lause
				if (jonoBotilta.onTyhj()) lisOmaLause();
			}

			// wait pttyy, kun aika menee umpeen tai uusiLause kutsuu notify, joten tss
			// kohdassa ei tiedet, pitk sanoa vai ksitell

			// ksitelln mahdollisesti odottava lause
			if (jonoBotille.onLauseita()) {
				// jos seuraava (ainoa) sanomista odottava lause on oma-aloitteinen,
				// poistetaan se uuden vastauksen tielt
				if (jonoBotilta.seuraavaOnOma()) jonoBotilta.seuraavaLause();

				// ksitelln seuraava lause jonosta (jota varten botti on hertetty)
				ksitteleSeuraavaLause();
			}
		}
	}

	/**
	 * Botin trkein rajapinta. Antaa botille uuden lauseen ksiteltvksi;
	 * lis lauseen ksittely odottavien lauseiden jonoon ja hertt botin
	 * ksittelemn lausetta synkronoidulla notifylla
	 *
	 * @param lause Uusi lause
	 */
	public void uusiLause(String lause)
	{
		// listn lause ksittely odottavien jonoon
		jonoBotille.lisLause(lause, System.currentTimeMillis(), true);

		// hertetn mahdollisesti nukkuva botti ksittelemn uutta lausetta
		synchronized (this) { notify(); }
	}

	/**
	 * Opettelee uusia sanoja tekstitiedostosta, eli parseroi tiedoston rivi kerrallaan
	 * aivan kuin tiedoston rivi olisi sanottu botille, ei vain vastaa mitn riveihin
	 *
	 * Luo uuden threadin opettelua varten, pyrii mukavasti taustalla eik kukaan jumiudu
	 *
	 * @param tiedosto Opeteltava tiedosto
	 * @return false jos virhe avattaessa tiedosto, muuten true
	 */
	public boolean opettele(final String tiedosto)
	{
		final Tiedostolukija t = new Tiedostolukija(tiedosto);

		if (!t.avattuOK()) return false;

		// tallennetaan sanojen yhteismr ja opittujen sanojen mr
		final int[] yhteens = {0, 0};

		new Thread(new Runnable() {
			public void run()
			{
				for (String s = t.seuraavaRivi(); s != null; s = t.seuraavaRivi()) {
					int[] parseroitu = parseri.parseroix(s);
					yhteens[0] += parseroitu[0];
					yhteens[1] += parseroitu[1];
				}

				t.sulje();

				// dumpataan statistiikat
				emo.ajattele("opeteltu '" + tiedosto + "' -> luettiin " + yhteens[1] + " (" +
				    yhteens[0] + ") [" + 100 * yhteens[1] / yhteens[0] + "%] uutta sanaa");
			}
		}).start();

		return true;
	}

	/**
	 * Opettelee uusia sanoja tekstitiedostosta, eli parseroi tiedoston rivi kerrallaan
	 * aivan kuin tiedoston rivi olisi sanottu botille, ei vain vastaa mitn riveihin
	 *
	 * (Jtetty tnne vain selkeyttvksi muistoksi, jumiuttaa kyttliittymn koska
	 * joudutaan odottamaan ett opettelu saadaan valmiiksi ennen kuin palataan)
	 *
	 * @param tiedosto Opeteltava tiedosto
	 * @return true jos kaikki meni kivasti; muuten false
	 */
	public boolean opetteleIlmanThreadia(String tiedosto)
	{
		Tiedostolukija t = new Tiedostolukija(tiedosto);

		if (!t.avattuOK()) return false;

		for (String s = t.seuraavaRivi(); s != null; s = t.seuraavaRivi())
			parseri.parseroi(s);

		return t.sulje();
	}

	/**
	 * Ksittelee seuraavan lauseen botille tulevien lauseiden jonosta
	 * ja lis vastauksen valmiiden lauseiden jonoon odottamaan ulostusta
	 */
	private void ksitteleSeuraavaLause()
	{
		int sanoja = parseri.parseroi(jonoBotille.seuraavaLause());

		lisVastausLause(sanoja);
	}

	/**
	 * Lis valmiiden lauseiden jonoon vastauksen botille sanottuun lauseeseen;
	 * vastauksen pituus saattaa (asetuksista riippuen) riippua sanotusta lauseesta
	 *
	 * @param sanotussaSanoja Kuinka monta sanaa botille sanotussa lauseessa oli
	 */
	private void lisVastausLause(int sanotussaSanoja)
	{
		// kuinka monta sanaa laitetaan listtvn lauseeseen
		int sanoja = Main.randomGauss(asetukset.lauseenPituusKA, asetukset.lauseenPituusKH);

		// laitetaanko lauseen pituus vastaamaan vastauksen pituutta
		if (Main.random() < asetukset.lauseenVastaavuusTN)
			// uusi (pituus) = arvottu + (vastattava - arvottu) * vastaavuussuhde
			sanoja += (sanotussaSanoja - sanoja) * asetukset.lauseenVastaavuusSuhde;

		// generoidaan satunnainen lause
		String lause = generoija.satunnainenLause(sanoja);

		// lasketaan lauseen odotettu sanomisaika (pituus / nopeus + edellisen sanomisaika)
		long aika = lause.length() * 1000
			    / Main.randomGauss(asetukset.kirjoitusnopeusKA, asetukset.kirjoitusnopeusKH)
			    + jonoBotilta.viimeisenAika();

		// listn lause botilta tulevien lauseiden jonoon vastauslauseena
		jonoBotilta.lisLause(lause, aika, true);
	}

	/**
	 * Lis valmiiden lauseiden jonoon satunnaisen, oma-aloitteisen lauseen;
	 * lauseen pituus riippuu vain asetuksissa mritellyist lausepituuksista
	 */
	private void lisOmaLause()
	{
		// kuinka monta sanaa laitetaan listtvn lauseeseen
		int sanoja = Main.randomGauss(asetukset.lauseenPituusKA, asetukset.lauseenPituusKH);

		// generoidaan satunnainen lause
		String lause = generoija.satunnainenLause(sanoja);

		// lasketaan lauseen odotettu sanomisaika (pituus / nopeus + edellisen sanomisaika)
		long aika = lause.length() * 1000
			    / Main.randomGauss(asetukset.kirjoitusnopeusKA, asetukset.kirjoitusnopeusKH)
			    + jonoBotilta.viimeisenAika();

		// listn laskettuun aikaan oma-aloitteisen lauseen odotusaika
		aika += Main.randomGauss(asetukset.satunnaisenLauseenOdotusKA,
					 asetukset.satunnaisenLuaseenOdotusKH) * 1000;

		// listn lause botilta tulevien lauseiden jonoon
		jonoBotilta.lisLause(lause, aika, false);
	}

	/**
	 * Sanoo (emon sano-metodilla) seuraavan lauseen valmiiden lauseiden jonosta
	 */
	private void sanoSeuraavaLause()
	{
		emo.sano(asetukset.bottiNick, jonoBotilta.seuraavaLause());
	}

	/**
	 * Sanoo (emon sano-metodilla) satunnaisen lauseen heti, ei lis sit jonoon
	 */
	private void sanoSatunnainenLause()
	{
		// kuinka monta sanaa laitetaan listtvn lauseeseen
		int sanoja = Main.randomGauss(asetukset.lauseenPituusKA, asetukset.lauseenPituusKH);

		// generoidaan satunnainen lause
		String lause = generoija.satunnainenLause(sanoja);

		emo.sano(asetukset.bottiNick, lause);
	}

}
